# -*- coding: utf-8 -*-
# Copyright (c) 2020-2022 Andreas Motl <andreas.motl@panodata.org>
# Copyright (c) 2020-2022 Richard Pobering <richard.pobering@panodata.org>
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
"""
Capture MQTT messages, using the `Paho MQTT Python Client`_, in the spirit of
`caplog` and `capsys`.

Source: https://github.com/hiveeyes/terkin-datalogger/blob/0.13.0/test/fixtures/mqtt.py

.. _Paho MQTT Python Client: https://github.com/eclipse/paho.mqtt.python
"""
import logging
import threading
import typing as t

import paho.mqtt.client as mqtt
import pytest

from pytest_mqtt.model import MqttMessage, MqttSettings
from pytest_mqtt.util import delay

logger = logging.getLogger(__name__)


class MqttClientAdapter(threading.Thread):
    def __init__(self, on_message_callback: t.Optional[t.Callable] = None, host: str = "localhost", port: int = 1883):
        super().__init__()
        self.client: mqtt.Client
        if not hasattr(mqtt, "CallbackAPIVersion"):
            # paho-mqtt 1.x
            self.client = mqtt.Client()
        else:
            # paho-mqtt 2.x
            self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION1)
        self.on_message_callback = on_message_callback
        self.host = host
        self.port = int(port)
        self.setup()

    def setup(self):
        client = self.client
        client.on_socket_open = self.on_socket_open
        client.on_connect = self.on_connect
        client.on_subscribe = self.on_subscribe
        client.on_message = self.on_message
        if self.on_message_callback:
            client.on_message = self.on_message_callback

        logger.debug("[PYTEST] Connecting to MQTT broker")
        client.connect(host=self.host, port=self.port)
        client.subscribe("#")

    def run(self):
        self.client.loop_start()

    def stop(self):
        logger.debug("[PYTEST] Disconnecting from MQTT broker")
        self.client.disconnect()
        self.client.loop_stop()

    def on_socket_open(self, client, userdata, sock):
        logger.debug("[PYTEST] Opened socket to MQTT broker")

    def on_connect(self, client, userdata, flags, rc):
        logger.debug("[PYTEST] Connected to MQTT broker")

    def on_subscribe(self, client, userdata, mid, granted_qos, properties=None):
        logger.debug("[PYTEST] Subscribed to MQTT topic(s)")

    def on_message(self, client, userdata, msg):
        logger.debug("[PYTEST] MQTT message received: %s", msg)

    def publish(self, topic: str, payload: str, **kwargs) -> mqtt.MQTTMessageInfo:
        message_info = self.client.publish(topic, payload, **kwargs)
        message_info.wait_for_publish()
        return message_info


class MqttCaptureFixture:
    """Provides access and control of log capturing."""

    def __init__(self, decode_utf8: t.Optional[bool], host: str = "localhost", port: int = 1883) -> None:
        """Creates a new funcarg."""
        self._buffer: t.List[MqttMessage] = []
        self._decode_utf8: bool = decode_utf8

        self.mqtt_client = MqttClientAdapter(on_message_callback=self.on_message, host=host, port=port)
        self.mqtt_client.start()
        # time.sleep(0.1)

    def on_message(self, client, userdata, msg):
        payload = msg.payload
        if self._decode_utf8:
            payload = payload.decode("utf-8")

        message = MqttMessage(
            topic=msg.topic,
            payload=payload,
            userdata=userdata,
        )
        self._buffer.append(message)
        logger.debug("[PYTEST] MQTT message received: %s", str(message)[:200])

    def finalize(self) -> None:
        """Finalizes the fixture."""
        self.mqtt_client.stop()
        self.mqtt_client.join(timeout=4.2)
        self._buffer = []

    @property
    def messages(self) -> t.List[MqttMessage]:
        return self._buffer

    @property
    def records(self) -> t.List[t.Tuple[str, t.Union[str, bytes], t.Dict]]:
        return [(item.topic, item.payload, item.userdata) for item in self._buffer]

    def publish(self, topic: str, payload: str, **kwargs) -> mqtt.MQTTMessageInfo:
        message_info = self.mqtt_client.publish(topic=topic, payload=payload, **kwargs)
        # Make the MQTT client publish and receive the message.
        delay()
        return message_info


@pytest.fixture(scope="function")
def capmqtt(request, mqtt_settings: MqttSettings):
    """Access and control MQTT messages."""

    # Configure `capmqtt` fixture, obtaining the `capmqtt_decode_utf8` setting from
    # either a global or module-wide setting, or from a test case marker.
    # https://docs.pytest.org/en/7.1.x/how-to/fixtures.html#fixtures-can-introspect-the-requesting-test-context
    # https://docs.pytest.org/en/7.1.x/how-to/fixtures.html#using-markers-to-pass-data-to-fixtures

    host, port = mqtt_settings.host, mqtt_settings.port

    capmqtt_decode_utf8 = (
        getattr(request.config.option, "capmqtt_decode_utf8", False)
        or getattr(request.module, "capmqtt_decode_utf8", False)
        or request.node.get_closest_marker("capmqtt_decode_utf8") is not None
    )
    result = MqttCaptureFixture(decode_utf8=capmqtt_decode_utf8, host=host, port=port)
    delay()
    yield result
    result.finalize()



# -*- coding: utf-8 -*-

# Copyright (c) 2020-2022 Andreas Motl <andreas.motl@panodata.org>
# Copyright (c) 2020-2022 Richard Pobering <richard.pobering@panodata.org>
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
"""
Provide the `Mosquitto`_ MQTT broker as a session-scoped fixture to your test
harness.

Source: https://github.com/hiveeyes/terkin-datalogger/blob/0.13.0/test/fixtures/mosquitto.py

.. _Mosquitto: https://github.com/eclipse/mosquitto
"""
import os

import docker
import pytest
from pytest_docker_fixtures import images
from pytest_docker_fixtures.containers._base import BaseImage

from pytest_mqtt.model import MqttSettings
from pytest_mqtt.util import probe_tcp_connect

images.settings["mosquitto"] = {
    "image": "eclipse-mosquitto",
    "version": "2.0.15",
    "options": {
        "command": "mosquitto -c /mosquitto-no-auth.conf",
        "publish_all_ports": False,
        "ports": {"1883/tcp": "1883"},
    },
}


class Mosquitto(BaseImage):

    name = "mosquitto"

    def __init__(self, host: str = "localhost", port: int = 1883) -> None:
        self.host = host
        self.port = port

    def check(self):
        # TODO: Add real implementation.
        return True

    def pull_image(self):
        """
        Image needs to be pulled explicitly.
        Workaround against `404 Client Error: Not Found for url: http+docker://localhost/v1.23/containers/create`.

        - https://github.com/jpmens/mqttwarn/pull/589#issuecomment-1249680740
        - https://github.com/docker/docker-py/issues/2101
        """
        docker_client = docker.from_env(version=self.docker_version)
        image_name = self.image
        docker_client.images.pull(image_name)

    def run(self):
        try:
            docker_client = docker.from_env(version=self.docker_version)
            docker_url = docker_client.api.base_url
        except Exception:
            raise ConnectionError("Cannot connect to the Docker daemon. Is the docker daemon running?")
        try:
            docker_client.ping()
        except Exception:
            raise ConnectionError(f"Cannot connect to the Docker daemon at {docker_url}. Is the docker daemon running?")
        self.pull_image()
        return super(Mosquitto, self).run()


mosquitto_image = Mosquitto()


def is_mosquitto_running(host: str, port: int) -> bool:
    return probe_tcp_connect(host, port)


def pytest_addoption(parser) -> None:
    parser.addoption("--mqtt-host", action="store", type=str, default="localhost", help="MQTT host name")
    parser.addoption("--mqtt-port", action="store", type=int, default=1883, help="MQTT port number")


@pytest.fixture(scope="session")
def mqtt_settings(pytestconfig) -> MqttSettings:
    return MqttSettings(
        host=pytestconfig.getoption("--mqtt-host"),
        port=pytestconfig.getoption("--mqtt-port"),
    )


@pytest.fixture(scope="session")
def mosquitto(mqtt_settings: MqttSettings):

    host, port = mqtt_settings.host, mqtt_settings.port

    # Gracefully skip spinning up the Docker container if Mosquitto is already running.
    if is_mosquitto_running(host, port):
        yield host, port
        return

    # Spin up Mosquitto container.
    if os.environ.get("MOSQUITTO"):
        yield os.environ["MOSQUITTO"].split(":")
    else:
        yield mosquitto_image.run()
        mosquitto_image.stop()



import socket
import threading


def delay(seconds: float = 0.075) -> None:
    """
    Wait for designated number of seconds.
    """
    threading.Event().wait(seconds)


def probe_tcp_connect(host: str, port: int) -> bool:
    """
    Test connecting to a remote TCP socket.
    https://github.com/lovelysystems/lovely.testlayers/blob/0.7.0/src/lovely/testlayers/util.py#L6-L13
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(0.1)
    ex = s.connect_ex((host, port))
    if ex == 0:
        s.close()
        return True
    return False



import dataclasses
import typing as t


@dataclasses.dataclass
class MqttMessage:
    """
    Container for `capmqtt`'s `message` response items.
    """

    topic: str
    payload: t.Union[str, bytes]
    userdata: t.Optional[t.Union[t.Dict, None]]


@dataclasses.dataclass
class MqttSettings:
    host: str
    port: int



try:
    from importlib.metadata import PackageNotFoundError, version  # noqa
except ImportError:  # pragma:nocover
    from importlib_metadata import PackageNotFoundError, version  # noqa

from .model import MqttMessage  # noqa:F401

try:
    __version__ = version("pytest-mqtt")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



#!/usr/bin/env python

# This is a shim to hopefully allow GitHub to detect the package.
# The build is done with Poetry.

import setuptools

if __name__ == "__main__":
    setuptools.setup(name="pytest-mqtt")
